// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//  Giuseppe Vacanti (cosine science & computing bv)
//  March 13, 2003
// 
//   $Id: Hkparameter.cc,v 1.3 2003/08/26 15:20:10 gvacanti Exp $

// ConcreteHkparameter
// A templated family of classes implementing the Hkparameter
// interface. It is not supposed to be used explicitly only via
// Hkparameter::make().

#include<Hkparameter.h>
#include<Hkparameter_Descriptor.h>
#include<Hkparameter_List.h>
#include<TimeRows.h>

#include<STime.h>
#include<DataFilter.h>
#include<Evaluator.h>
#include<DataReader.h>
#include<SmartDal>

#include<algorithm>
#include<numeric>
#include<sstream>

template<class T> class ConcreteHkparameter:
  public virtual Hkparameter {
public:
  ConcreteHkparameter(const Hkparameter_Descriptor & id, unsigned hktolerate)
    : hkparameter_(id), is_set_(false), tolerate_(hktolerate) {
    if(!Evaluator::isValidExpression(id.condition))
      errstr << "The condition '" << id.condition
	     << "' associated with the housekeeping parameter "
	     << id.name << " is not a valid expression."
	     << Sas::error("InvalidSelectionExpression");
  }
  std::string name() const { return hkparameter_.name; }
  std::string colname() const { return hkparameter_.colname; }
  std::string condition() const { return hkparameter_.condition; }
  std::string description() const { return hkparameter_.description; }
  Xmm::Instrument instrument() const { return hkparameter_.instrument; }
  Xmm::DataMode hktype() const { return hkparameter_.hktype; }
  bool periodic() const { return hkparameter_.periodic; }
  std::string value() const;
  void process(const SmartDal::TableH & table, const STime & start, const STime & stop);
  bool is_set() const { return is_set_; }
  
private:
  const Hkparameter_Descriptor hkparameter_;
  T value_;
  bool is_set_;
  unsigned long tolerate_;
  DataReader<T> dr_;

  void acquire(const SmartDal::TableH & table, const STime & start, const STime & stop);
  void evaluate();

};

template<class T> void ConcreteHkparameter<T>::acquire(const SmartDal::TableH & table,
						       const STime & start,
						       const STime & stop){

  // First filter on the condition. Do this because the condtion is
  // meant to be used to disentangle the mess that the periodic HK
  // files are. For two reasons:

  // [1] Non periodic HK files have entries with the same time tag
  //    TIME                CCDID WINDOWX0
  //   2000-01-01T00:00:00    1    10
  //   2000-01-01T00:00:00    2    10
  //   2000-01-01T00:00:00    3    10
  //   2000-01-01T00:01:00    1    10
  //   2000-01-01T00:01:00    2    10
  //   2000-01-01T00:01:00    3    10
  // In order to be able to find the correct entry for CCD 1 we have
  // first to filter on CCD=1, then find the time.

  // [2] Records applying to the same exposure need not come with the
  // same time tag. That is, this is possible:
  //    TIME                CCDID WINDOWX0
  //   2000-01-01T00:00:00    1    10
  //   2000-01-01T00:00:10    2    10
  //   2000-01-01T00:00:10    3    10
  //   2000-01-01T00:01:00    1    10
  //   2000-01-01T00:01:00    2    10
  //   2000-01-01T00:01:00    3    10
  // Now here if the exposure start is at 00:00:30, the active range
  // filter (below) would return only the second and third record. If
  // I then apply the condition about CCD==1, no records are selected.

  // For these reasons, one must first filter on the condition
  // (CCD==1), and then on the time.

  std::string expression = condition();
  DataFilter df(expression);
  // Row ranges returned here are counted from 0.
  std::vector<DataFilter::RowRange> rr;
  df.doSelect(*table.get(), rr); // selectlib does not know about
				 // SmartDal objects, so I use the
				 // get() method.
  if(rr.size() != 0){
    errstr << table->qualifiedName() << ": the condition " << expression
	   << " identifies the following row ranges for inclusion:\n";
    for(unsigned k = 0; k < rr.size(); ++k)
      errstr << rr[k].startRow << " to " << rr[k].stopRow << endl;
    errstr << Sas::message(Msg::App, Msg::Noisy);
  }

  TimeRows tr(table,rr);
  rr.clear();
  long low, upp;

  // The range of rows to look at is different depending on whether
  // the parameter is periodic or not.
  if(periodic()){
    low = tr.index(tr.lower_bound(start));
    upp = tr.index(tr.upper_bound(stop));
  } else {
    std::pair<long, long> p = tr.active_range(start);
    low = p.first;
    upp = p.second;
  }

  std::stringstream s;  
  // Perverse!! The DAL starts from 0 but in selectlib #ROW starts from 1.
  unsigned from = low + 1;
  unsigned to = upp + 1;
  s << condition() + " && (#ROW >= " << from << " && #ROW < " << to << ")";
  expression = s.str();
  DataFilter df2(expression);
  df2.doSelect(*table.get(), rr); // selectlib does not know about
                                  // SmartDal objects, so I use the
                                  // get() method.

  // Rewrite the condition so that when we report the results the row
  // numbers start all from 0.
  s.str("");
  s << condition() + " && (#ROW >= " << low << " && #ROW < " << upp << ")";  
  expression = s.str();

  // There can only be one range of consecutive rows.  
  if(rr.size() == 0)
    errstr << "In " + table->qualifiedName() + " the condition " << expression
	   << " selects no rows." << Sas::error("NoUsableRangeOfEntries");
  else if(rr.size() > 1){
    errstr << "In " + table->qualifiedName() + " the condition " << expression
	   << " selects more than one row range:\n";
    for(size_t k = 0; k < rr.size(); ++k)
      errstr << "range " << k << " = [" << rr[k].startRow
	     << ", " << rr[k].stopRow << "]\n";
    errstr << Sas::error("NoUsableRangeOfEntries");
  } else
    errstr << "In " + table->qualifiedName() + " the condition " << expression
	   << " selects the row range [" << rr[0].startRow
	   << ", " << rr[0].stopRow << "]"
	   << Sas::message(Msg::App, Msg::Noisy);

  

  // Seek the table so that only the useful range is read by dr_
  table->seek(rr[0].startRow, rr[0].stopRow - rr[0].startRow + 1);
  dr_(table->column(colname()));
  // But 'un-seek' it because most functions in other parts of the SAS
  // do not expect to see a table with a seek.
  table->seek(0, table->rows());
}  

// Here we have different bahavior depending on the type of the
// parameter. Real parameters will be averaged, the others will be
// checkd to see if they are constant. In the latter case a number of
// mismatches can be accepted.

template<class T> void ConcreteHkparameter<T>::evaluate(){
  std::vector<T> & vd = dr_.data();  
  // Value constant. Check must happen over range of rows selected
  // above, and here we assume that dr_ has been filled only with the
  // values we're supposed to check.  This algorithm does not show at
  // which row the value is not constant. A number of discrepant
  // values can be accepted, depending on the configuration: it has
  // happened that just one or two values out of 100's were different.
  value_ = vd[0];

  unsigned long count_different = vd.size() - std::count(vd.begin(), vd.end(), value_);
  if(count_different == 0)
    is_set_ = true;
  else if(count_different <= tolerate_){
    errstr << "While looking at the housekeeping parameter "
	   << name() << "[column=" << colname() << ", condition="
	   << condition() << ", instrument=" << Xmm::toString(instrument())
	   << ", periodic=" << periodic() << "] I found " << count_different
	   << " entries out of " << vd.size() << " to be different from the expected value " << value_ << ". I'm configured to tolerate up to " << tolerate_ << " discrepant entries." << Sas::message(Msg::App, Msg::Silent);
    is_set_ = true;
  } else {
    errstr << "While looking at the housekeeping parameter "
	   << name() << "[column=" << colname() << ", condition="
	   << condition() << ", instrument=" << Xmm::toString(instrument())
	   << ", periodic=" << periodic() << "] I found " << count_different
	   << " entries out of " << vd.size() << " to be different from the expected value " << value_ << ". This is more than I'm configured to accept (" << tolerate_ << ")." << Sas::error("HkparameterNotConstant");
    is_set_ = false;
  }
}

// FIXME: can reduce duplication?

template<> void ConcreteHkparameter<real32>::evaluate(){
  std::vector<real32> & vd = dr_.data();
  real32 sum = 0.;
  sum = std::accumulate(vd.begin(), vd.end(), sum);
  value_ = sum/vd.size();
  is_set_ = true;
}
template<> void ConcreteHkparameter<real64>::evaluate(){
  std::vector<real64> & vd = dr_.data();
  real64 sum = 0.;
  sum = std::accumulate(vd.begin(), vd.end(), sum);
  value_ = sum/vd.size();
  is_set_ = true;
}
  
template<class T> void ConcreteHkparameter<T>::process(const SmartDal::TableH & table,
						       const STime & start,
						       const STime & stop)
{

  acquire(table, start, stop);
  evaluate();
}

// Here we convert the internal value of type T to a string. The int8
// speciallization is needed because int8 is typedef-ed to be char,
// and without an explicit cast it is printed as character instead of
// integer.
template<class T> std::string ConcreteHkparameter<T>::value() const {
  if(!is_set_)
    errstr << "I was asked to return the value of the housekeeping parameter "
	   << name() << "[column=" << colname() << ", condition="
	   << condition() << ", instrument=" << Xmm::toString(instrument())
	   << ", periodic=" << periodic() << "] but no value has been set."
	   << Sas::error("ValueNotSet");
  std::stringstream ss;
  ss << value_;
  return ss.str();
}
template<> std::string ConcreteHkparameter<int8>::value() const {
  if(!is_set_)
    errstr << "I was asked to return the value of the housekeeping parameter "
	   << name() << "[column=" << colname() << ", condition="
	   << condition() << ", instrument=" << Xmm::toString(instrument())
	   << ", periodic=" << periodic() << "] but no value has been set."
	   << Sas::error("ValueNotSet");
  std::stringstream ss;
  ss << static_cast<unsigned>(value_);
  return ss.str();
}

Hkparameter * Hkparameter::make(const Hkparameter_Descriptor & id, unsigned hktolerate) {

  switch(id.dtype) {
  case Dal::Int8:
    return new ConcreteHkparameter<int8>(id, hktolerate);
    break;
  case Dal::Int16:
    return new ConcreteHkparameter<int16>(id, hktolerate);
    break;
  case Dal::Uint16:
    return new ConcreteHkparameter<uint16>(id, hktolerate);
    break;
  case Dal::Int32:
    return new ConcreteHkparameter<int32>(id, hktolerate);
    break;
  case Dal::Uint32:
    return new ConcreteHkparameter<uint32>(id, hktolerate);
    break;
  case Dal::Real32:
    return new ConcreteHkparameter<real32>(id, hktolerate);
    break;
  case Dal::Real64:
    return new ConcreteHkparameter<real64>(id, hktolerate);
    break;
  case Dal::DString:
    return new ConcreteHkparameter<std::string>(id, hktolerate);
    break;
  default:
    Sas::error("NoHkparameterType", "");
  }

  return 0;
}

std::vector<Hkparameter *>  Hkparameter::get_parameters(Xmm::Instrument instrument, unsigned hktolerate){
  std::vector<Hkparameter *> v;
  const unsigned n = sizeof(hkd) /  sizeof(Hkparameter_Descriptor);
  for(unsigned k = 0; k < n ; ++k)
    if(hkd[k].instrument == instrument)
      v.push_back(Hkparameter::make(hkd[k], hktolerate));
  return v;
}
    
