// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// June 2000
//

#include <fstream>
#include <string>
// ios_base should be in <ios> 
//#include <ios> 
#include <param.h>
#include <errstr.h>
#include <Tracer.h>
#include <OdfObject.h>
#include <OalServer.h>
#include <Xmm.h>
#include <SasDalUtils.h>

#include "odfingest.h"

/*
 * (SOC-SCR-214) Due to a problem in the ODF generation, some ODFs
 * have the ATS EXTNAME keyword changed to "SCATS49" in stead of 
 * "SCATS1".
 */
static void check_time_range(const string& table, string extName, const STime& start,
			     const STime& stop, const string& tname = "TIME")
{
  PROC
  DataSet *set;
  set = dataSetServer->open(table, Dal::Read);

  if (extName == ":SCATS1")
    if (!set->hasBlock("SCATS1"))
      extName=":SCATS49";

  Column *c = set->table(table+extName)->column(tname);
  ColumnData *cd = c->data();
  STime t0(SasDalUtils::getStringColumnRow(cd, 0));
  STime tlast(SasDalUtils::getStringColumnRow(cd, c->rows() - 1));
  if(t0 > start || tlast < stop)
    errstr << "Time range for " << set->qualifiedName()
	   << ": " << t0 << " to " << tlast
	   << " does not fully contain the observation span "
	   << start << " to " << stop
	   << Sas::warning("InconsistentTimeRange");
  delete cd;
  dataSetServer->close(set);
  ENDPROC
}

void odfingest () {
  PROC

  string odfdir;
  if(booleanParameter("withodfdir"))
    odfdir = stringParameter("odfdir");
  else
    odfdir =  oalServer.odf()->directoryName();

  OdfObject o = OdfObject(odfdir,
			  booleanParameter("findinstrumentmodes"),
			  booleanParameter("usehousekeeping"));

  if(o.valid()){
    ofstream os;
    string name;
    if(booleanParameter("usecanonicalname")){
      os.open((stringParameter("outdir") + "/"
	       + o.canonicalSummaryName()).c_str());
      name = o.canonicalSummaryName();
      // Make name available to calling meta-tasks
      setParameter("summaryfile", name);
    } else {
      os.open(stringParameter("summaryfile").c_str());
      string s = stringParameter("summaryfile");
      string::size_type p = s.find_last_of("/");
      if(p != string::npos)
	name = s.substr(p);
      else
	name = s;
    }
    // FIXME: use ios::failure
    if(!os)
      Sas::error("FileIo", "A file i/o error occurred while creating the ODF summary file " + name + ". Checks: you have write permission; you are not trying to overwrite a summary file in use.");
    o.insertFile(name);
    o.removeFile(o.icdosf());
    o.writeSummary(os, booleanParameter("writepath"));
    
    // FIXME: use ios::failure
    if(!os)
      Sas::error("FileIo", "A file i/o error occurred while creating the ODF summary file " + name + ". Checks: you have write permission; you are not trying to overwrite a summary file in use.");
    os.close();
    

    if(booleanParameter("oalcheck")){
      const Odf *odf = oalServer.newOdf(name);
      STime obsstart = odf->observationStartTime();
      STime obsstop = odf->observationEndTime();
      
      errstr << "Directory: " << odf->directoryName() << endl;
      errstr << "Revolution: " << odf->orbitNr() << " Proposal identifier: "
	     << odf->proposalId() << endl;
      errstr << "Observation start time: " << obsstart << endl;
      errstr << "Observation stop time: " << obsstop << endl;
      const ProposalInfo pi = odf->proposalInfo();
      errstr << "Target: " << pi.targetName << " (" << pi.targetRa << ","
	     << pi.targetDec << ")" << endl;
      errstr << "Pointing: " << pi.boresightRa << " " << pi.boresightDec << endl;
      errstr << "Position angle constraints: " << "(" << pi.posAngleConstLo
	     << " to " << pi.posAngleConstHi << ") " << pi.posAngleOrigRef
	     << Sas::message(Msg::App, Msg::Sparse);

      if(!odf->hasSet(Xmm::SC, Xmm::ATTITUDE_HISTORY))
	errstr << "Missing attitude history file"
	       << Sas::warning("MissingOdfComponent");
      else
	check_time_range(odf->selectSet(Xmm::SC, Xmm::ATTITUDE_HISTORY)->name(),
			 ":SCATS1", obsstart, obsstop, "VALTIME");      

      if(!odf->hasSet(Xmm::SC, Xmm::TIME_CORRELATION))
	errstr << "Missing time correlation file"
	       << Sas::warning("MissingOdfComponent");
      else
	check_time_range(odf->selectSet(Xmm::SC, Xmm::TIME_CORRELATION)->name(),
			 ":SCTCS1", obsstart, obsstop, "ERTIME");      

      if(!odf->hasSet(Xmm::SC, Xmm::RECONSTRUCTED_ORBIT)){
	errstr << "Missing reconstructed orbit file"
	       << Sas::warning("MissingOdfComponent");
	if(!odf->hasSet(Xmm::SC, Xmm::PREDICTED_ORBIT))
	  errstr << "Missing predicted orbit file"
		 << Sas::warning("MissingOdfComponent");
      
      }
    }
  } else
    errstr << "No valid ODF found in " << odfdir
	   << Sas::error("NoOdfFound");      
  
  ENDPROC
}
